#' fit_rfci_lu2019

#' fit_rfci_lu2019
#' Adaptation of method from L2019 with Ranger
#' See fit_lu2019() for more details
#' @param dataset dataset with target and predictors
#' @param target column name of the target
#' @param predictors column name of the predictors
#' @param n.cores n.cores
#' @keywords cipred
#' @export
#' @examples
#' \dontrun{
#' fit_rfci_lu2019()
#' }
#
fit_rfci_lu2019 <- function(dataset,target,predictors,n.cores=1){

  res <- fit_lu2019(dataset=dataset,target=target,predictors=predictors,n.cores=n.cores)

  return(res)
}

#' pred_rfci_lu2019_prime

#' pred_rfci_lu2019_prime
#' Adaptation of method from L2019 with Ranger
#' See predict_lu2019() for more details
#' @param lu2019 lu2019 model only
#' @param dataset dataset with target and predictors
#' @param target column name of the target
#' @param predictors column name of the predictors
#' @keywords cipred
#' @export
#' @examples
#' \dontrun{
#' pred_rfci_lu2019_prime()
#' }
#
pred_rfci_lu2019_prime <- function(lu2019,dataset,target,predictors){

  tmp <- predict_lu2019_prime(lu2019=lu2019,
                      X.test=as.data.frame(dataset)[,predictors],
                      Y.test = as.data.frame(dataset)[,target],
                      alpha = 0.05,
                      n.cores = 1)

  res <- data.frame(obs=tmp$obs,
                    prediction=tmp$pred,
                    lower_pred_0_05 = tmp$'lower_pred_0.05',
                    upper_pred_0_05 = tmp$'upper_pred_0.05',
                    reu_lower_pred_0_05 = tmp$'rel_lower_pred_0.05',
                    reu_upper_pred_0_05 = tmp$'rel_upper_pred_0.05',
                    pvalue        = tmp$p_value_obs,
                    method          = rep("ranger_lu2019_prim",length(tmp$obs)))

  return(res)
}


#' pred_rfci_lu2019

#' pred_rfci_lu2019
#' Adaptation of method from L2019 with Ranger
#' See predict_lu2019() for more details
#' @param lu2019 lu2019 model only
#' @param dataset dataset with target and predictors
#' @param predictors column name of the predictors
#' @keywords cipred
#' @export
#' @examples
#' \dontrun{
#' pred_rfci_lu2019()
#' }
#
pred_rfci_lu2019 <- function(lu2019,dataset,predictors){

  tmp <- predict_lu2019(lu2019=lu2019,
                      X.test=as.data.frame(dataset)[,predictors],
                      alpha = 0.05,
                      n.cores = 1)

  res <- data.frame(prediction=tmp$pred,
                    lower_pred_0_05 = tmp$'lower_pred_0.05',
                    upper_pred_0_05 = tmp$'upper_pred_0.05',
                    reu_lower_pred_0_05 = tmp$'rel_lower_pred_0.05',
                    reu_upper_pred_0_05 = tmp$'rel_upper_pred_0.05',
                    method          = rep("ranger_lu2019",length(tmp$obs)))

  return(res)
}

# code adapted from function quantForestError() in package forestError,
# https://github.com/benjilu/forestError/blob/master/R/quantforesterror.R
# Like quantForestError(), predict_lu2019() provides the prediction and the quantiles reprenstation of Q_alpha
# In addition, the function provides a relative Q_alpha against the observation and the p-value of the observation against the errors related to the specific tree/nodes

# avoid check note
if(getRversion() >= "2.15.1"){utils::globalVariables(c("tree", "terminal_node",
                                                     ".", "oob_error", "rowid_test", "pred","obs",
                                                     "node_errs", "pred", "all_errs", "..col_names",".I",
                                                     "..p_col_names", "..q_col_names"))}

#' fit_lu2019

#' fit_lu2019
#' code adapted from function quantForestError() in package forestError,
#' https://github.com/benjilu/forestError/blob/master/R/quantforesterror.R
#' It first connects the terminal nodes used by the input while predicting to the ones used while training the algorithm.
#' It then gathers the out of bag of prediction error from the training related to these specific input.
#' The prediction is then connected to the ensemble of out of bag of prediction errors from the training.
#' @param dataset that gather target and predictors as a matrix or a data.frame
#' @param target colname of the target
#' @param predictors colname of the predictors
#' @param n.cores nb of cores
#' @keywords cipred
#' @export
#' @examples
#' \dontrun{
#' fit_lu2019()
#' }
#
fit_lu2019 <- function(dataset,target,predictors,n.cores = 1){

  forest <- fit_ranger(dataset=dataset,target=target,predictors=predictors)

  # get terminal nodes of all observations
  train.terminal.nodes <- stats::predict(forest, dataset[,predictors], num.threads = n.cores, type = "terminalNodes")$predictions

  # get number of times each training observation appears in each tree
  bag.count <- matrix(unlist(forest$inbag.counts, use.names = FALSE), ncol = forest$num.trees, byrow = FALSE)

  # get the OOB prediction error of each training observation
  oob.errors <- dataset[,target] - forest$predictions

  # get the terminal nodes of the training observations in the trees in which
  # they are OOB (for all other trees, set the terminal node to be NA)
  train.terminal.nodes[bag.count != 0] <- NA

  # reshape train.terminal.nodes to be a long data.table and include OOB
  # prediction errors as a column
  long_train_nodes <- data.table::as.data.table(train.terminal.nodes)
  long_train_nodes[, `:=`(oob_error = oob.errors)]
  long_train_nodes <- data.table::melt(
    long_train_nodes,
    id.vars = c("oob_error"),
    measure.vars = 1:ncol(train.terminal.nodes),
    variable.name = "tree",
    value.name = "terminal_node",
    variable.factor = FALSE,
    na.rm = TRUE)

  # collapse the long data.table by unique tree/node
  long_train_nodes <- long_train_nodes[,
    .(node_errs = list(oob_error)),
    keyby = c("tree", "terminal_node")]

  res <- list(model = forest,
              long_train_nodes = long_train_nodes)
  class(res) <- c("lu2019",class(res))

  return(res)
}

#' predict_lu2019_prime

#' predict_lu2019_prime
#' code adapted from function quantForestError() in package forestError,
#' https://github.com/benjilu/forestError/blob/master/R/quantforesterror.R
#' Algorithm predictive conditional error distribution
#' The prediction is first connected to the ensemble of out of bag of prediction errors from the training.
#' The interval is then processed by picking-up the value of the sorted ensemble at the rank corresponding to a specific percentile.**
#' @param lu2019 Model of class lu2019 that combine Ranger RF model and long train nodes
#' @param X.test Predictors used for the prediction , as a matrix or a data.frame
#' @param Y.test Target to be compared to the prediction
#' @param alpha type-I error rates desired for the conditional prediction intervals
#' @param n.cores nb of cores
#' @keywords cipred
#' @export
#' @examples
#' \dontrun{
#' predict_lu2019_prime()
#' }
#
predict_lu2019_prime <- function(lu2019, X.test, Y.test, alpha, n.cores = 1) {

  if (! inherits(lu2019,"lu2019")) stop("object lu2019 must be of class lu2019")

  forest <- lu2019$model
  long_train_nodes <- lu2019$long_train_nodes

  # get terminal nodes of all observations
  test.terminal.nodes <- stats::predict(forest, X.test, num.threads = n.cores, type = "terminalNodes")$predictions

  # get test observation predictions
  test.preds <- stats::predict(forest, X.test)$predictions

  # reshape test.terminal.nodes to be a long data.table and
  # add unique IDs and predicted values
  long_test_nodes <- data.table::melt(
    data.table::as.data.table(test.terminal.nodes)[, `:=`(rowid_test = .I, pred = test.preds, obs = Y.test)],
    id.vars = c("rowid_test", "pred","obs"),
    measure.vars = 1:ncol(test.terminal.nodes),
    variable.name = "tree",
    variable.factor = FALSE,
    value.name = "terminal_node")

  # set key columns for faster indexing
  data.table::setkey(long_test_nodes, tree, terminal_node)

  # join the train and test edgelists by tree/node and compute relevant summary
  # statistics via chaining
  col_names <- c("rowid_test", "pred","obs","all_errs")
  oob_error_stats <-
    long_train_nodes[long_test_nodes,
                     .(tree, terminal_node, rowid_test, pred, obs, node_errs)][,
                                                                          .(all_errs = list(sort(unlist(node_errs)))),
                                                                          keyby = c("rowid_test", "pred","obs")]

  col_names <- c("pred","obs")

  #################################
  # Q-value alpha from all_errs   #
  #################################
  percentiles <- sort(c(alpha / 2, 1 - (alpha / 2)))
  interval_col_names <- paste0(rep(c("lower_pred_", "upper_pred_"), each = length(alpha)),
                               c(alpha, rev(alpha)))
  col_names <- c(col_names, interval_col_names)

  # compute Q-values alpha
  oob_error_stats[, (interval_col_names) := lapply(percentiles, FUN = function(p) {
    pred + purrr::map_dbl(all_errs, ~.x[ceiling(length(.x) * p)])})]

  ##########################################
  # Relative Q-value alpha from all_errs   #
  ##########################################
  percentiles <- sort(c(alpha / 2, 1 - (alpha / 2)))
  rel_interval_col_names <- paste0(rep(c("rel_lower_pred_", "rel_upper_pred_"), each = length(alpha)),
                               c(alpha, rev(alpha)))
  col_names <- c(col_names, rel_interval_col_names)

  # compute Q-values alpha
  oob_error_stats[, (rel_interval_col_names) := lapply(percentiles, FUN = function(p) {
    100*1/pred * purrr::map_dbl(all_errs, ~.x[ceiling(length(.x) * p)])})]

  #####################################
  # P-value of obs against all_errs   #
  #####################################
  pred_window_col_names <- "pred_window"
  col_names <- c(col_names, pred_window_col_names)

  tmp <-purrr::map(oob_error_stats[, all_errs],~.x)

  pred_window <- lapply(1:nrow(oob_error_stats),function(y){
     res <- oob_error_stats[y, pred] + tmp[[y]]
  })

  oob_error_stats[, (pred_window_col_names)] <- pred_window

  p_value_col_names <- "p_value_obs"
  col_names <- c(col_names, p_value_col_names)

  tmp <-purrr::map(oob_error_stats[, pred_window],~.x)

  p_value_obs <- lapply(1:nrow(oob_error_stats),function(y){
     res <- mean(tmp[[y]]<=oob_error_stats[y, obs],na.rm=TRUE)
  })

  oob_error_stats[, (p_value_col_names)] <- unlist(p_value_obs)


  #results
  col_names <- c("pred","obs",interval_col_names,rel_interval_col_names,p_value_col_names)
  output = data.table::setDF(oob_error_stats[, ..col_names])

  return(output)
}

#' predict_lu2019

#' predict_lu2019
#' code adapted from function quantForestError() in package forestError,
#' https://github.com/benjilu/forestError/blob/master/R/quantforesterror.R
#' Algorithm predictive conditional error distribution
#' The prediction is first connected to the ensemble of out of bag of prediction errors from the training.
#' The interval is then processed by picking-up the value of the sorted ensemble at the rank corresponding to a specific percentile.**
#' @param lu2019 Model of class lu2019 that combine Ranger RF model and long train nodes
#' @param X.test Predictors used for the prediction , as a matrix or a data.frame
#' @param alpha type-I error rates desired for the conditional prediction intervals
#' @param n.cores nb of cores
#' @keywords cipred
#' @export
#' @examples
#' \dontrun{
#' predict_lu2019()
#' }
#
predict_lu2019 <- function(lu2019, X.test, alpha, n.cores = 1) {

  if (! inherits(lu2019,"lu2019")) stop("object lu2019 must be of class lu2019")

  forest <- lu2019$model
  long_train_nodes <- lu2019$long_train_nodes

  # get terminal nodes of all observations
  test.terminal.nodes <- stats::predict(forest, X.test, num.threads = n.cores, type = "terminalNodes")$predictions

  # get test observation predictions
  test.preds <- stats::predict(forest, X.test)$predictions

  # reshape test.terminal.nodes to be a long data.table and
  # add unique IDs and predicted values
  long_test_nodes <- data.table::melt(
    data.table::as.data.table(test.terminal.nodes)[, `:=`(rowid_test = .I, pred = test.preds)],
    id.vars = c("rowid_test", "pred"),
    measure.vars = 1:ncol(test.terminal.nodes),
    variable.name = "tree",
    variable.factor = FALSE,
    value.name = "terminal_node")

  # set key columns for faster indexing
  data.table::setkey(long_test_nodes, tree, terminal_node)

  # join the train and test edgelists by tree/node and compute relevant summary
  # statistics via chaining
  col_names <- c("rowid_test", "pred","all_errs")
  oob_error_stats <-
    long_train_nodes[long_test_nodes,
                     .(tree, terminal_node, rowid_test, pred, node_errs)][,
                                                                          .(all_errs = list(sort(unlist(node_errs)))),
                                                                          keyby = c("rowid_test", "pred")]

  col_names <- c("pred")

  #################################
  # Q-value alpha from all_errs   #
  #################################
  percentiles <- sort(c(alpha / 2, 1 - (alpha / 2)))
  interval_col_names <- paste0(rep(c("lower_pred_", "upper_pred_"), each = length(alpha)),
                               c(alpha, rev(alpha)))
  col_names <- c(col_names, interval_col_names)

  # compute Q-values alpha
  oob_error_stats[, (interval_col_names) := lapply(percentiles, FUN = function(p) {
    pred + purrr::map_dbl(all_errs, ~.x[ceiling(length(.x) * p)])})]

  ##########################################
  # Relative Q-value alpha from all_errs   #
  ##########################################
  percentiles <- sort(c(alpha / 2, 1 - (alpha / 2)))
  rel_interval_col_names <- paste0(rep(c("rel_lower_pred_", "rel_upper_pred_"), each = length(alpha)),
                               c(alpha, rev(alpha)))
  col_names <- c(col_names, rel_interval_col_names)

  # compute Q-values alpha
  oob_error_stats[, (rel_interval_col_names) := lapply(percentiles, FUN = function(p) {
    100*1/pred * purrr::map_dbl(all_errs, ~.x[ceiling(length(.x) * p)])})]


  #results
  col_names <- c("pred",interval_col_names,rel_interval_col_names)
  output <- data.table::setDF(oob_error_stats[, ..col_names])

  return(output)
}
